module AMS; end
module AMS::ConvexDecomposition
  class << self

    # One 2D convexhull algorithm by me.
    # @param [Array<Geom::Point3d, Array<Numeric>>] points
    # @return [Array<Geom::Point3d> An array of outer points.
    # @raise [TypeError] if points size is less than three.
    def get_outer_points(points)
      points = points.clone
      # Turn all points to Point3d objects and remove similar points.
      for i in 0...points.size
        next if points[i].nil?
        points[i] = Geom::Point3d.new(points[i].x, points[i].y, 0)
        for j in 0...points.size
          next if j == i || points[j].nil?
          points[j] = nil if points[j] == points[i]
        end
      end
      # Verify unique points size.
      if points.size < 3
        raise(TypeError, "Expected at least 3 unique points, but got #{points.size} unique points.", caller)
      elsif points.size == 3
        return points
      end
      # Remove all inner and collinear points.
      for i in 0...points.size
        next if points[i].nil?
        init_theta = nil
        min_theta = 0
        max_theta = 0
        for j in 0...points.size
          next if j == i || points[j].nil?
          dir = points[i].vector_to(points[j]).normalize
          theta = Math.acos(dir.x)
          theta = Math::PI*2 - theta if dir.y < 0
          unless init_theta
            init_theta = theta
            next
          end
          theta -= init_theta
          # Make range from [0, 360).
          theta += Math::PI*2 if theta < 0
          theta -= Math::PI*2 if theta > Math::PI*2
          # Make the difference angle from (-180, 180]
          theta -= Math::PI*2 if theta > Math::PI
          min_theta = theta if theta < min_theta
          max_theta = theta if theta > max_theta
          # Remove point if difference is greater than or equal to 180 degrees.
          if max_theta - min_theta >= Math::PI
            points[i] = nil
            break
          end
        end
      end
      points.compact!
      points
    end

    def exact_2d_convex_decomposition(face)
      unless face.is_a?(Sketchup::Face)
        raise(TypeError, "Expected Sketchup::Face, but got #{face.class}.", caller)
      end
      model = Sketchup.active_model
      group = model.entities.add_group
      group.entities.add_faces_from_mesh(face.mesh, 0)
      tra = Geom::Transformation.new(ORIGIN, face.normal)
      group.entities.grep(Sketchup::Edge).each { |edge|
        next if !edge.valid? || edge.faces.size != 2
        verts = (edge.faces[0].vertices + edge.faces[1].vertices).uniq
        b_convex = true
        for i in 0...verts.size
          init_theta = nil
          min_theta = 0
          max_theta = 0
          for j in 0...verts.size
            next if j == i
            p1 = verts[i].position.transform(tra.inverse)
            p1.z = 0
            p2 = verts[j].position.transform(tra.inverse)
            p2.z = 0
            dir = p1.vector_to(p2).normalize
            theta = Math.acos(dir.x)
            theta = Math::PI*2 - theta if dir.y < 0
            unless init_theta
              init_theta = theta
              next
            end
            theta -= init_theta
            # Make range from [0, 360).
            theta += Math::PI*2 if theta < 0
            theta -= Math::PI*2 if theta > Math::PI*2
            # Make the difference angle from (-180, 180]
            theta -= Math::PI*2 if theta > Math::PI
            min_theta = theta if theta < min_theta
            max_theta = theta if theta > max_theta
            if max_theta - min_theta > Math::PI
              b_convex = false
              break
            end
          end
          break unless b_convex
        end
        edge.erase! if b_convex
      }
      group
    end

  end # class << self
end # module AMS::ConvexDecomposition

unless file_loaded?(__FILE__)
  file_loaded(__FILE__)
  menu = UI.menu('Plugins')
  item = menu.add_item('AMS 2D Convex Decomposition'){
    model = Sketchup.active_model
    model.start_operation('AMS 2D Convex Decomposition', true)
    model.selection.grep(Sketchup::Face).each { |e|
      AMS::ConvexDecomposition.exact_2d_convex_decomposition(e)
    }
    model.commit_operation
  }
  menu.set_validation_proc(item){
    Sketchup.active_model.selection.grep(Sketchup::Face).size > 0 ? MF_ENABLED : MF_GRAYED
  }
end
